"""Helper functions for fetching system objects."""

from collections.abc import AsyncIterator, Sequence
from contextlib import suppress
from typing import Any

from aiovantage.errors import ClientError, ClientResponseError

from . import ConfigClient
from .interfaces.configuration import (
    CloseFilter,
    GetFilterResults,
    GetObject,
    OpenFilter,
)
from .interfaces.introspection import GetVersion


async def get_objects(
    client: ConfigClient,
    *,
    types: Sequence[str] | None = None,
    xpath: str | None = None,
) -> AsyncIterator[Any]:
    """Get all vantage system objects of the specified types.

    Args:
        client: The config client instance
        types: An optional string, or list of strings of object types to fetch
        xpath: An optional xpath to filter the results by, eg. "/Load", "/*[@VID='12']"

    Yields:
        The objects of the specified types
    """
    # Support both a single object type and a list of status types
    if isinstance(types, str):
        types = [types]
    elif types is not None:
        types = list(types)

    # Open the filter
    handle = await client.request(
        OpenFilter, OpenFilter.Params(object_types=types, xpath=xpath)
    )

    if handle is None:
        raise ClientResponseError("Failed to open filter")

    try:
        # Get the results
        while True:
            response = await client.request(
                GetFilterResults, GetFilterResults.Params(h_filter=handle)
            )

            if not response:
                break

            for obj in response:
                if obj.choice is None:
                    continue

                yield obj.choice
    finally:
        # Close the filter
        with suppress(ClientError):
            await client.request(CloseFilter, handle)


async def get_objects_by_id(
    client: ConfigClient, vids: Sequence[int]
) -> AsyncIterator[Any]:
    """Get all vantage system objects of the specified ids.

    Args:
        client: The config client instance
        vids: A list of Vantage IDs for object to fetch

    Yields:
        The objects of the specified ids
    """
    # Open the filter
    response = await client.request(GetObject, GetObject.Params(vids=list(vids)))
    if not response:
        return

    for obj in response:
        if obj.choice is None:
            continue

        yield obj.choice


async def get_object_by_id(client: ConfigClient, vid: int) -> Any:
    """Get a single Vantage system object by id.

    Args:
        client: The config client instance
        vid: The Vantage ID of the object to fetch

    Returns:
        The object matching the specified id, or None if not found
    """
    try:
        return await get_objects_by_id(client, [vid]).__anext__()
    except StopAsyncIteration:
        return None


async def get_version(client: ConfigClient) -> str | None:
    """Get the firmware version of the Vantage controller.

    Args:
        client: The config client instance.
    """
    version = await client.request(GetVersion)
    if version is None:
        return None

    return version.app



"""Base class for custom device objects."""

from dataclasses import dataclass

from .location_object import LocationObject


@dataclass
class CustomDevice(LocationObject):
    """Base class for custom device objects."""



"""ModuleGen2 object, eg. SDM12-EM, UDM08-EM."""

from dataclasses import dataclass, field

from .system_object import SystemObject
from .types import Parent


@dataclass
class ModuleGen2(SystemObject):
    """ModuleGen2 object, eg. SDM12-EM, UDM08-EM."""

    parent: Parent = field(
        metadata={
            "name": "Parent",
        }
    )



"""Controller holding and managing Vantage buttons."""

from typing_extensions import override

from aiovantage.command_client.object_interfaces import ButtonInterface
from aiovantage.models import Button

from .base import BaseController


class ButtonsController(BaseController[Button], ButtonInterface):
    """Controller holding and managing Vantage buttons."""

    vantage_types = ("Button",)
    """The Vantage object types that this controller will fetch."""

    status_types = ("BTN",)
    """Which Vantage 'STATUS' types this controller handles, if any."""

    @override
    async def fetch_object_state(self, vid: int) -> None:
        """Fetch the state properties of a dry contact."""
        state = {
            # Buttons are momentary, default to not pressed to avoid a lookup
            "pressed": False,
        }

        self.update_state(vid, state)

    @override
    def handle_status(self, vid: int, status: str, *args: str) -> None:
        """Handle simple status messages from the event stream."""
        if status != "BTN":
            return

        # STATUS BTN
        # -> S:BTN <id> <state (PRESS/RELEASE)>
        state = {
            "pressed": args[0] == "PRESS",
        }

        self.update_state(vid, state)



"""Controller holding and managing Vantage anemo (wind) sensors."""

from decimal import Decimal

from typing_extensions import override

from aiovantage.command_client.object_interfaces import (
    AnemoSensorInterface,
    SensorInterface,
)
from aiovantage.models import AnemoSensor

from .base import BaseController


class AnemoSensorsController(
    BaseController[AnemoSensor], AnemoSensorInterface, SensorInterface
):
    """Controller holding and managing Vantage anemo (wind) sensors."""

    vantage_types = ("AnemoSensor",)
    """The Vantage object types that this controller will fetch."""

    status_types = ("WIND",)
    """Which Vantage 'STATUS' types this controller handles, if any."""

    @override
    async def fetch_object_state(self, vid: int) -> None:
        """Fetch the state properties of an anemo sensor."""
        state = {
            "speed": await AnemoSensorInterface.get_speed(self, vid),
        }

        self.update_state(vid, state)

    @override
    def handle_status(self, vid: int, status: str, *args: str) -> None:
        """Handle simple status message from the event stream."""
        if status != "WIND":
            return

        # STATUS WIND
        # -> S:WIND <id> <wind_speed>
        state = {
            "speed": Decimal(args[0]),
        }

        self.update_state(vid, state)



"""Set the value of a variable on the Vantage controller."""

import argparse
import asyncio
import contextlib
import logging

from aiovantage import Vantage

# Grab connection info from command line arguments
parser = argparse.ArgumentParser(description="aiovantage example")
parser.add_argument("host", help="hostname of Vantage controller")
parser.add_argument("id", help="GMem id to set")
parser.add_argument("value", help="value to set GMem to")
parser.add_argument("--username", help="username for Vantage controller")
parser.add_argument("--password", help="password for Vantage controller")
parser.add_argument("--debug", help="enable debug logging", action="store_true")
args = parser.parse_args()


async def main() -> None:
    """Run code example."""
    if args.debug:
        logging.basicConfig(level=logging.DEBUG)

    # Connect to the Vantage controller
    async with Vantage(args.host, args.username, args.password) as vantage:
        # Look up the object by id
        try:
            gmem_id = int(args.id)
            gmem = await vantage.gmem.aget(gmem_id)
            if gmem is None:
                print(f"GMem object with id '{gmem_id}' does not exist")
                return
        except ValueError:
            print(f"Couldn't parse object id '{args.id}'")
            return

        # Convert the value to the correct type
        if gmem.is_bool:
            if args.value.lower() in ("true", "1", "on"):
                args.value = True
            elif args.value.lower() in ("false", "0", "off"):
                args.value = False
            else:
                print("Invalid boolean value")
                return
        elif gmem.is_int:
            try:
                args.value = int(args.value)
            except ValueError:
                print("Invalid integer value")
                return

        # Set the value
        print(f"Setting '{gmem.name}' (id = {gmem.id}) to '{args.value}'")
        await vantage.gmem.set_value(gmem_id, args.value)


with contextlib.suppress(KeyboardInterrupt):
    asyncio.run(main())



"""Controller holding and managing Vantage power profiles."""

from aiovantage.models import PowerProfile

from .base import BaseController


class PowerProfilesController(BaseController[PowerProfile]):
    """Controller holding and managing Vantage power profiles."""

    vantage_types = ("PowerProfile", "DCPowerProfile", "PWMPowerProfile")
    """The Vantage object types that this controller will fetch."""
